import configparser
import copy
import os.path
from configparser import ConfigParser
from io import StringIO
from textwrap import dedent

import pytest

from configupdater import (
    AlreadyAttachedError,
    AssignMultilineValueError,
    Comment,
    ConfigUpdater,
    DuplicateOptionError,
    DuplicateSectionError,
    MissingSectionHeaderError,
    NoConfigFileReadError,
    NoneValueDisallowed,
    NoOptionError,
    NoSectionError,
    NotAttachedError,
    Option,
    ParsingError,
    Section,
    Space,
)


def test_reade_write_no_changes(setup_cfg_path, setup_cfg):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    assert str(updater) == setup_cfg


def test_read_file_like(setup_cfg_path):
    updater = ConfigUpdater()
    with open(setup_cfg_path) as fp:
        updater.read_file(fp)
    fp = StringIO()
    updater.read_file(fp)


def test_update_no_changes(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    old_mtime = os.path.getmtime(setup_cfg_path)
    updater.update_file()
    new_mtime = os.path.getmtime(setup_cfg_path)
    assert old_mtime != new_mtime


def test_updater_to_dict(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    parser = ConfigParser()
    parser.read(setup_cfg_path)
    parser_dict = {sect: dict(parser[sect]) for sect in parser.sections()}
    assert updater.to_dict() == parser_dict


def test_update_no_cfg():
    updater = ConfigUpdater()
    with pytest.raises(NoConfigFileReadError):
        updater.update_file()


def test_update_file_issue_68(tmp_path):
    # allow_no_value = False
    file = tmp_path / "file.cfg"
    file.write_text("[section]")
    cfg = ConfigUpdater(allow_no_value=False).read_file(open(file))
    cfg.set("section", "option")
    with pytest.warns(NoneValueDisallowed):
        # A warning is issued, but the method will not fail
        cfg.update_file()
    assert file.read_text().strip() == "[section]"

    # allow_no_value = True
    file.write_text("[section]")
    cfg = ConfigUpdater(allow_no_value=True).read_file(open(file))
    cfg.set("section", "option")
    cfg.update_file()
    assert file.read_text().strip() == "[section]\noption"


def test_str(setup_cfg_path, setup_cfg):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    output = str(updater)
    assert output == setup_cfg


def test_has_section(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    assert updater.has_section("metadata")
    assert not updater.has_section("nonexistent_section")


def test_contains_section(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    assert "metadata" in updater


def test_sections(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    exp_sections = [
        "metadata",
        "options",
        "options.packages.find",
        "options.extras_require",
        "test",
        "tool:pytest",
        "aliases",
        "bdist_wheel",
        "build_sphinx",
        "devpi:upload",
        "flake8",
        "pyscaffold",
    ]
    assert updater.sections() == exp_sections


def test_len_updater(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    # we test the number of blocks, not sections
    assert len(updater) == 14


def test_iter_section(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    # iter_blocks will give us everything, iter_sections just the section objects and
    # __iter__ should work as iter_blocks
    assert len([block for block in updater.iter_blocks()]) == 14
    assert len([block for block in updater.iter_sections()]) == 12
    assert len([entry for entry in updater]) == 12


def test_iter_items_section(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)

    result = list(k for k, v in updater["metadata"].items())
    assert result == [
        "name",
        "description",
        "author",
        "author-email",
        "license",
        "url",
        "long-description",
        "platforms",
        "classifiers",
    ]


def test_get_section(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    section = updater["metadata"]
    assert section._structure
    with pytest.raises(KeyError):
        updater["non_existent_section"]

    # `get_section` is equivalent to dict.get
    assert updater.get_section("non_existent_section") is None
    default = {}
    assert updater.get_section("non_existent_section", default) is default


def test_section_to_dict(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    parser = ConfigParser()
    parser.read(setup_cfg_path)
    updater_dict = updater["metadata"].to_dict()
    parser_dict = dict(parser["metadata"])
    assert updater_dict == parser_dict


def test_has_option(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    assert updater.has_option("metadata", "author")
    assert not updater.has_option("metadata", "coauthor")
    assert not updater.has_option("nonexistent_section", "key")


def test_contains_options(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    section = updater["metadata"]
    assert "author" in section
    assert "wrong_option" not in section


def test_len_section(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    section = updater["metadata"]
    # we test the number of blocks in section
    assert len(section) == 12


def test_len_option(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    option = updater["metadata"]["classifiers"]
    assert len(option.lines) == 3


def test_iter_option(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    section = updater["metadata"]
    # iter_blocks will give us everything, iter_options just the option objects and
    # __iter__ should work as iter_blocks
    assert len([entry for entry in section.iter_blocks()]) == 12
    assert len([entry for entry in section.iter_options()]) == 9
    assert len([entry for entry in section]) == 9


def test_get_options(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    options = updater.options("options.packages.find")
    exp_options = ["where", "exclude"]
    assert options == exp_options
    with pytest.raises(NoSectionError):
        updater.options("non_existent_section")

    # Section objects also have a "get" method similar to dict
    section = updater["metadata"]
    assert section.get("name").value == "configupdater"
    assert section.get("non_existent_option") is None
    assert section.get("non_existent_option", []) == []


def test_items(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    sect_names, sects = zip(*updater.items())
    exp_sect_namess = [
        "metadata",
        "options",
        "options.packages.find",
        "options.extras_require",
        "test",
        "tool:pytest",
        "aliases",
        "bdist_wheel",
        "build_sphinx",
        "devpi:upload",
        "flake8",
        "pyscaffold",
    ]
    assert list(sect_names) == exp_sect_namess
    assert all([isinstance(s, Section) for s in sects])

    opt_names, opts = zip(*updater.items("devpi:upload"))
    exp_opt_names = ["no-vcs", "formats"]
    assert list(opt_names) == exp_opt_names
    assert all([isinstance(o, Option) for o in opts])


def test_get_method(setup_cfg_path):
    updater = ConfigUpdater()
    updater.read(setup_cfg_path)
    value = updater.get("metadata", "license").value
    assert value == "mit"
    with pytest.raises(NoSectionError):
        updater.get("non_existent_section", "license")
    with pytest.raises(NoOptionError):
        updater.get("metadata", "wrong_key")
    assert updater.get("metadata", "wrong_key", fallback=None) is None


test1_cfg_in = """
[default]
key = 1
"""

test1_cfg_out = """
[default]
key = 2
"""

test1_cfg_out_added = """
[default]
key = 1
other_key = 3
"""

test1_cfg_out_none = """
[default]
key
"""

test1_cfg_out_removed = """
[default]
"""

test1_cfg_out_values = """
[default]
key =
    param1
    param2
"""


def test_get_option():
    updater = ConfigUpdater()
    updater.read_string(test1_cfg_in)
    updater.read_string(test1_cfg_in)
    option = updater["default"]["key"]
    assert option.value == "1"
    with pytest.raises(KeyError):
        updater["default"]["wrong_key"]


def test_value_change():
    updater = ConfigUpdater()
    updater.read_string(test1_cfg_in)
    assert updater["default"]["key"].value == "1"
    updater["default"]["key"].value = "2"
    assert str(updater) == test1_cfg_out
    # try another way
    section = updater["default"]
    section["key"] = "1"
    assert str(updater) == test1_cfg_in


def test_remove_option():
    updater = ConfigUpdater()
    updater.read_string(test1_cfg_in)
    updater.remove_option("default", "key")
    assert str(updater) == test1_cfg_out_removed
    updater.remove_option("default", "non_existent_key")
    updater.read_string(test1_cfg_in)
    del updater["default"]["key"]
    assert str(updater) == test1_cfg_out_removed
    with pytest.raises(NoSectionError):
        updater.remove_option("wrong_section", "key")


def test_set_option():
    updater = ConfigUpdater(allow_no_value=True)
    updater.read_string(test1_cfg_in)
    updater.set("default", "key", "1")
    assert updater["default"]["key"].value == "1"
    updater.set("default", "key", 2)
    assert updater["default"]["key"].value == 2
    assert str(updater) == test1_cfg_out
    updater.set("default", "key")
    assert updater["default"]["key"].value is None
    assert str(updater) == test1_cfg_out_none
    updater.read_string(test1_cfg_in)
    updater.set("default", "other_key", 3)
    assert str(updater) == test1_cfg_out_added
    updater.read_string(test1_cfg_in)
    values = ["param1", "param2"]
    updater["default"]["key"].set_values(values)
    assert str(updater) == test1_cfg_out_values
    assert values == ["param1", "param2"]  # non destructive operation
    with pytest.raises(NoSectionError):
        updater.set("wrong_section", "key", "1")
    new_option = copy.deepcopy(updater["default"]["key"])
    updater["default"]["key"] = new_option
    assert updater["default"]["key"] is new_option
    new_option = copy.deepcopy(updater["default"]["key"])
    new_option.key = "wrong_key"
    with pytest.raises(ValueError):
        updater["default"]["key"] = new_option


def test_section_set_option():
    updater = ConfigUpdater(allow_no_value=True)
    updater.read_string(test1_cfg_in)
    default_sec = updater["default"]
    default_sec.set("key", "1")
    assert default_sec["key"].value == "1"
    default_sec.set("key", 2)
    assert default_sec["key"].value == 2
    assert str(default_sec) == test1_cfg_out[1:]
    default_sec.set("key")
    assert default_sec["key"].value is None
    assert str(default_sec) == test1_cfg_out_none[1:]
    updater.read_string(test1_cfg_in)
    default_sec = updater["default"]
    default_sec.set("other_key", 3)
    assert str(default_sec) == test1_cfg_out_added[1:]
    updater.read_string(test1_cfg_in)
    default_sec = updater["default"]
    values = ["param1", "param2"]
    default_sec["key"].set_values(values)
    assert str(default_sec) == test1_cfg_out_values[1:]
    assert values == ["param1", "param2"]  # non destructive operation


def test_del_option():
    updater = ConfigUpdater()
    updater.read_string(test1_cfg_in)
    del updater["default"]["key"]
    assert str(updater) == "\n[default]\n"
    with pytest.raises(KeyError):
        del updater["default"]["key"]


test2_cfg_in = """
[section1]
key = 1

[section2]
key = 2
"""

test2_cfg_out = """
[section1]
key = 1

"""


def test_del_section():
    updater = ConfigUpdater()
    updater.read_string(test2_cfg_in)
    del updater["section2"]
    assert str(updater) == test2_cfg_out
    with pytest.raises(KeyError):
        del updater["section2"]


test_wrong_cfg = """
[strange section]
a
"""


def test_handle_error():
    updater = ConfigUpdater(allow_no_value=False)
    with pytest.raises(ParsingError):
        updater.read_string(test_wrong_cfg)


def test_validate_format(setup_cfg_path):
    updater = ConfigUpdater(allow_no_value=True)
    updater.read(setup_cfg_path)
    updater.validate_format(allow_no_value=False)
    updater.set("metadata", "author")
    with pytest.raises(ParsingError):
        updater.validate_format(allow_no_value=False)


def test_validate_format_propagates_kwargs(setup_cfg_path):
    updater = ConfigUpdater(allow_no_value=True)
    updater.read(setup_cfg_path)
    updater.set("pyscaffold", "namespace")
    assert updater["pyscaffold"]["namespace"].value is None
    assert str(updater["pyscaffold"]["namespace"]) == "namespace\n"  # no `=` sign
    assert updater.validate_format() is True


test3_cfg_in = """
[section]
key = 1
"""

test3_cfg_out = """
# comment of section
[section]
key = 1
# comment after section
"""


def test_add_before_after_comment():
    updater = ConfigUpdater()
    updater.read_string(test3_cfg_in)
    updater["section"].add_before.comment("comment of section")
    updater["section"].add_after.comment("# comment after section\n")
    assert str(updater) == test3_cfg_out


test4_cfg_in = """
[section]
key1 = 1
"""

test4_cfg_out = """
[section]
key0 = 0
key1 = 1
key2
"""


def test_add_before_after_option():
    updater = ConfigUpdater(allow_no_value=True)
    updater.read_string(test4_cfg_in)
    with pytest.raises(ValueError):
        updater["section"].add_before.option("key0", 0)
    updater["section"]["key1"].add_before.option("key0", 0)
    updater["section"]["key1"].add_after.option("key2")
    assert str(updater) == test4_cfg_out


test5_cfg_in = """
[section0]
key0 = 0
[section1]
key1 = 1
key2 = 2
"""


test5_cfg_out = """
[section0]
key0 = 0


[section1]
key1 = 1

key2 = 2
"""


def test_add_before_after_space():
    updater = ConfigUpdater()
    updater.read_string(test5_cfg_in)
    updater["section1"].add_before.space(2)
    updater["section1"]["key1"].add_after.space(1)
    assert str(updater) == test5_cfg_out


test6_cfg_in = """
[section0]
key0 = 0
[section2]
key1 = 1
key2 = 2
"""


test6_cfg_out = """
[section0]
key0 = 0
[section1]
key1 = 42
[section2]
key1 = 1
key2 = 2

[section3]
"""


test6_cfg_out_new_sect = """
[section0]
key0 = 0
[section2]
key1 = 1
key2 = 2
[new_section]
key0 = 0
"""


def test_add_before_after_section():
    updater = ConfigUpdater()
    updater.read_string(test6_cfg_in)
    with pytest.raises(ValueError):
        updater["section2"]["key1"].add_before.section("section1")
    updater["section2"].add_before.section("section1")
    updater["section1"]["key1"] = 42
    updater["section2"].add_after.space(1).section("section3")
    assert str(updater) == test6_cfg_out
    with pytest.raises(ValueError):
        updater["section2"].add_before.section(updater["section2"]["key1"])
    updater.read_string(test6_cfg_in)
    sect_updater = ConfigUpdater()
    sect_updater.read_string(test6_cfg_in)
    section = sect_updater["section0"]
    section.name = "new_section"
    updater["section2"].add_after.section(section.detach())
    assert str(updater) == test6_cfg_out_new_sect
    with pytest.raises(DuplicateSectionError):
        updater["section2"].add_after.section(section)


test7_cfg_in = """
[section0]
key0 = 0

[section1]
key1 = 42
"""


test7_cfg_out = """
[section0]
key0 = 0

[section1]
key1 = 42
[section2]
key1 = 1
"""


def test_add_section():
    updater = ConfigUpdater()
    updater.read_string(test7_cfg_in)
    with pytest.raises(DuplicateSectionError):
        updater.add_section("section1")
    updater.add_section("section2")
    updater["section2"]["key1"] = 1
    assert str(updater) == test7_cfg_out
    with pytest.raises(ValueError):
        updater.add_section(updater["section2"]["key1"])


test6_cfg_out_overwritten = """
[section0]
key0 = 42
[section2]
key1 = 1
key2 = 2
"""


def test_set_item_section():
    updater = ConfigUpdater()
    sect_updater = ConfigUpdater()
    updater.read_string(test6_cfg_in)
    with pytest.raises(ValueError):
        updater["section"] = "newsection"
    sect_updater.read_string(test6_cfg_in)
    section = sect_updater["section0"]
    with pytest.raises(ValueError):
        updater["new_section"] = section
    section.name = "new_section"
    updater["new_section"] = section.detach()
    assert str(updater) == test6_cfg_out_new_sect
    # test overwriting an existing section
    updater.read_string(test6_cfg_in)
    sect_updater.read_string(test6_cfg_in)
    exist_section = sect_updater["section0"]
    exist_section["key0"] = 42
    updater["section0"] = exist_section.detach()
    assert str(updater) == test6_cfg_out_overwritten


def test_no_space_around_delim():
    updater = ConfigUpdater(space_around_delimiters=False)
    updater.read_string(test7_cfg_in)
    updater["section0"]["key0"] = 0
    del updater["section1"]
    assert str(updater) == "\n[section0]\nkey0=0\n\n"


def test_constructor(setup_cfg_path):
    updater = ConfigUpdater(delimiters=(":", "="))
    updater.read(setup_cfg_path)
    updater = ConfigUpdater(delimiters=(":", "="), allow_no_value=True)
    updater.read(setup_cfg_path)


test8_inline_prefixes = """
[section] # just a section
key = value  # just a value
"""


def test_inline_comments():
    updater = ConfigUpdater(inline_comment_prefixes="#")
    updater.read_string(test8_inline_prefixes)
    assert updater.has_section("section")
    assert updater["section"]["key"].value == "value"


test9_dup_section = """
[section]
key = value

[section]
key = value
"""


def test_duplicate_section_error():
    updater = ConfigUpdater()
    with pytest.raises(DuplicateSectionError):
        updater.read_string(test9_dup_section)


def test_missing_section_error():
    updater = ConfigUpdater()
    with pytest.raises(MissingSectionHeaderError):
        updater.read_string("key = value")


test10_dup_option = """
[section]
key = value
key = value
"""


def test_duplicate_option_error():
    updater = ConfigUpdater()
    with pytest.raises(DuplicateOptionError):
        updater.read_string(test10_dup_option)


test11_no_values = """
[section]
key
"""


def test_no_value():
    updater = ConfigUpdater(allow_no_value=True)
    updater.read_string(test11_no_values)
    assert updater["section"]["key"].value is None


def test_eq(setup_cfg_path):
    updater1 = ConfigUpdater()
    updater1.read(setup_cfg_path)
    updater2 = ConfigUpdater()
    updater2.read(setup_cfg_path)
    assert updater1 == updater2
    updater1.remove_section("metadata")
    assert updater1 != updater2
    assert updater1 != updater2["metadata"]
    assert updater2["metadata"] != updater2["metadata"]["author"]
    assert not updater1.remove_section("metadata")


test12_cfg_in = """
[section]
opiton = 42
"""


test12_cfg_out = """
[section]
option = 42
"""


def test_rename_option_key():
    updater = ConfigUpdater()
    updater.read_string(test12_cfg_in)
    updater["section"]["opiton"].key = "option"


test13_cfg_in = """
[section]
CAPITAL = 1
"""


def test_set_optionxform():
    updater = ConfigUpdater()
    updater.read_string(test13_cfg_in)
    assert updater["section"]["capital"].value == "1"
    assert callable(updater.optionxform)
    updater.optionxform = lambda x: x
    updater.read_string(test13_cfg_in)
    assert updater["section"]["CAPITAL"].value == "1"


test14_cfg_in = """
[section]
option2 = 2
option4 = 4
"""

test14_cfg_out = """
[section]
option0 = 0
option1 = 1
option2 = 2
option3 = 3
option4 = 4
"""


def test_insert_at():
    updater = ConfigUpdater()
    updater.read_string(test14_cfg_in)
    updater["section"].insert_at(0).option("option0", 0).option("option1", 1)
    updater["section"].insert_at(3).option("option3", 3)
    assert str(updater) == test14_cfg_out


def test_read_file_with_string():
    updater = ConfigUpdater()
    with pytest.raises(RuntimeError):
        updater.read_file("path/to/file.cfg")


test15_cfg_in = """
[section]
OptionA = 2
"""


def test_read_mixed_case_options():
    updater = ConfigUpdater()
    updater.read_string(test15_cfg_in)
    assert updater.has_option("section", "OptionA")
    assert updater.has_option("section", "optiona")
    assert updater["section"]["OptionA"].value == "2"
    assert updater["section"]["optiona"].value == "2"


test16_cfg_in = """
[section]
OptionA = 2
"""


test16_cfg_out = """
[section]
OptionA = 4
OptionB = 6
"""


def test_update_mixed_case_options():
    updater = ConfigUpdater()
    updater.read_string(test16_cfg_in)
    updater["section"]["optiona"].value = "4"
    updater["section"]["OptionB"] = "6"
    assert str(updater) == test16_cfg_out


test17_cfg_in = """
[section]
key1 = 1
"""


test17_cfg_out = """
[section]
key0 = 0
key1 = 1
key2 = 2
key3 = 3
"""


def test_add_before_then_add_after_option():
    updater = ConfigUpdater()
    updater.read_string(test17_cfg_in)
    updater["section"]["key1"].add_before.option("key0", "0")
    updater["section"]["key1"].add_after.option("key2", "2")
    updater["section"]["key2"].add_after.option("key3", "3")
    assert str(updater) == test17_cfg_out


test18_cfg_in = """
[section]
Key0 = 0
"""

test18_cfg_out = """
[section]
Key0 = 2
"""


def test_assure_no_duplicate_options():
    updater = ConfigUpdater()
    updater.read_string(test18_cfg_in)
    updater["section"]["KEY0"].value = "1"
    updater["section"]["keY0"] = "2"
    assert str(updater) == test18_cfg_out


test19_cfg_in = """
[section]
Key0 = 0
"""


def test_no_duplicate_blocks_with_blockbuilder():
    updater = ConfigUpdater()
    updater.read_string(test19_cfg_in)
    with pytest.raises(DuplicateOptionError):
        updater["section"]["Key0"].add_after.option("key0", "1")
    with pytest.raises(DuplicateSectionError):
        updater["section"].add_after.section("section")
    assert str(updater) == test19_cfg_in


# Taken from issue #14
test20_cfg_in = """
[flake8]
exclude =
  # Trash and cache:
  .git
  __pycache__
  .venv
  .eggs
  *.egg
  temp
  # Bad code that I write to test things:
  ex.py
new = value

per-file-ignores =
  # Disable imports in `__init__.py`:
  lambdas/__init__.py: WPS226, WPS413
  lambdas/contrib/mypy/lambdas_plugin.py: WPS437
  # There are multiple assert's in tests:
  tests/*.py: S101, WPS226, WPS432, WPS436, WPS450
  # We need to write tests to our private class:
  tests/test_math_expression/*.py: S101, WPS432, WPS450"""


def test_comments_in_multiline_options():
    updater = ConfigUpdater()
    updater.read_string(test20_cfg_in)
    per_file_ignores = updater["flake8"]["per-file-ignores"].value
    exp_val = (
        "\n# Disable imports in `__init__.py`:\nlambdas/__init__.py: WPS226, WPS413\n"
        "lambdas/contrib/mypy/lambdas_plugin.py: WPS437\n# There are multiple assert's"
        " in tests:\ntests/*.py: S101, WPS226, WPS432, WPS436, WPS450\n# We need to"
        " write tests to our private class:\ntests/test_math_expression/*.py: S101,"
        " WPS432, WPS450"
    )
    updater.validate_format()
    assert per_file_ignores == exp_val
    assert test20_cfg_in == str(updater)


test21_cfg_in = """
[main1]
key1 =
    a
    b

    c
    d
[main2]
key1 =
    a
    b

    c
    d

[main3]

[main4]
key1 =
    a
    b


    c
    d
key2 =
    # comment
    a
    b
    # comment


    c
    d
[main5]
key1 =
    a
    b

    c
    d

key2 = abcd
[main6]
"""


def test_empty_lines_in_values_support():
    updater = ConfigUpdater()
    updater.read_string(test21_cfg_in)
    parser = ConfigParser()
    parser.read_string(test21_cfg_in)
    assert updater["main1"]["key1"].value == parser["main1"]["key1"]
    assert updater["main2"]["key1"].value == parser["main2"]["key1"]
    assert updater["main4"]["key1"].value == parser["main4"]["key2"]

    # Make sure values are not being stored in "whitespace nodes"
    value = (x.strip() for x in updater["main1"]["key1"].value.splitlines())
    assert [v for v in value if v] == ["a", "b", "c", "d"]

    assert test21_cfg_in == str(updater)
    with pytest.raises(ParsingError):
        updater = ConfigUpdater(empty_lines_in_values=False)
        updater.read_string(test21_cfg_in)


test22_cfg_in = """
[section]
key1 = 1
# comment
key2 = 2

"""

test22_cfg_out = """
[section]
key1 = 1
key2 = 2
"""


def test_navigation_and_remove():
    updater = ConfigUpdater()
    updater.read_string(test22_cfg_in)
    section = updater["section"]
    key1 = section["key1"]
    assert key1 is updater["section"].first_block
    assert key1.previous_block is None
    assert isinstance(key1.next_block, Comment)
    key2 = key1.next_block.next_block
    assert key2.value == section["key2"].value
    assert isinstance(key2.next_block, Space)
    assert key2.next_block is section.last_block
    assert section.last_block.next_block is None
    key1.next_block.detach()
    key2.next_block.detach()
    assert str(updater) == test22_cfg_out


test23_cfg_in = """
[section1]
    # indented comment
a = 42
"""


def test_indented_comment():
    updater = ConfigUpdater()
    updater.read_string(test23_cfg_in)
    section = updater["section1"]
    assert isinstance(section.first_block, Comment)
    assert str(updater) == test23_cfg_in


test24_cfg_in = """
[sec1]
key1 = val1
key2 = val2

[sec2]
key1 = val1
key2 = val2
"""


def test_iter_consistency_with_configparser():
    parser = ConfigParser()
    parser.read_string(test24_cfg_in)
    updater = ConfigUpdater()
    updater.read_string(test24_cfg_in)
    # [1:] to drop 'DEFAULT' section
    assert list(parser)[1:] == list(updater)
    assert list(parser["sec1"]) == list(updater["sec1"])


def test_add_detached_section_option_objects():
    updater = ConfigUpdater()
    updater.read_string(test24_cfg_in)
    sec1 = updater["sec1"]
    sec2 = updater["sec2"]
    assert sec2.container is updater
    sec2.detach()
    assert not updater.has_section("sec2")
    assert not sec2.has_container()
    with pytest.raises(NotAttachedError):
        sec2.container

    new_sec2 = Section("new-sec2")
    new_opt = Option(key="new-key", value="new-value")
    new_sec2.add_option(new_opt)
    with pytest.raises(AlreadyAttachedError):
        sec1.add_option(new_opt)
    updater.add_section(new_sec2)
    assert updater.has_section("new-sec2")
    assert updater["new-sec2"]["new-key"].value == "new-value"

    new_sec3 = Section("new-sec3")
    new_opt2 = Option(key="new-key", value="new-value")
    updater["new-sec3"] = new_sec3
    new_sec3["new-key"] = new_opt2


def test_transferring_blocks_between_elements():
    # Let's say a user have a big new section that it wants to insert in an existing
    # document. Instead of programmatically creating this new section using the builder
    # API, they might be tempted to parse this new section from template files.

    existing = """\
    [section0]
    option0 = 0
    """

    template1 = """\
    [section1]
    option1 = 1
    """

    template2 = """\
    [section2]
    option2 = 2
    # comment
    """

    target = ConfigUpdater()
    target.read_string(dedent(existing))

    source1 = ConfigUpdater()
    source1.read_string(dedent(template1))

    source2 = ConfigUpdater()
    source2.read_string(dedent(template2))

    target["section1"] = source1["section1"].detach()
    assert "section1" in target

    target["section1"].add_after.section(source2["section2"].detach())
    assert "section2" not in source1
    assert "section2" in target


def test_transfering_sections_and_manipulating_options():
    # This test case is similar to `test_transfering_sections_between_elements`,
    # but now we use the builder API in the options inside the transferred section.
    # We need to make sure that when a section is copied to a different tree,
    # the references inside the options are updated, so we don't end up adding blocks in
    # the original object.

    existing = """\
    [section0]
    option0 = 0
    """

    template1 = """\
    [section1]
    option1 = 1
    """

    target = ConfigUpdater()
    target.read_string(dedent(existing))

    source1 = ConfigUpdater()
    source1.read_string(dedent(template1))

    target["section1"] = source1["section1"].detach()
    assert "section1" in target

    target["section1"]["option1"].add_after.option("option2", "value")

    assert "option2" not in str(source1)
    assert "option2" in str(target)


def test_transfering_sections_and_manipulating_options_with_deepcopy():
    existing = """\
    [section0]
    option0 = 0
    """

    template1 = """\
    [section1]
    option1 = 1
    """

    template2 = """\
    [section2]
    option2 = 2
    # comment
    """

    target = ConfigUpdater()
    target.read_string(dedent(existing))

    source1 = ConfigUpdater()
    source1.read_string(dedent(template1))

    source2 = ConfigUpdater()
    source2.read_string(dedent(template2))

    target["section1"] = copy.deepcopy(source1["section1"])
    assert "section1" in target
    assert "section1" in source1

    target["section1"].add_after.section(copy.deepcopy(source2["section2"]))
    target["section2"]["option2"] = "value"
    assert "option2 = value" in str(target)
    assert "option2" not in str(source1)
    assert "value" not in str(source2)


def test_deepcopy():
    example = """\
    [options]
    testing =   # Add here test requirements (used by tox)
        sphinx  # required for system tests
        flake8  # required for system tests
    """
    updater = ConfigUpdater().read_string(dedent(example))
    clone = copy.deepcopy(updater)
    assert clone["options"].container is clone
    assert clone["options"]["testing"].container is not updater["options"]
    clone["options"]["testing"].detach()
    assert "testing" not in str(clone)
    assert "testing" in str(updater)
    clone.add_section("other")
    assert "other" in clone
    assert "other" not in updater
    assert str(updater) == dedent(example)


def test_section_comment():
    section = """\
    [section0] # section comment
    key = value
    """

    updater = ConfigUpdater()
    updater.read_string(dedent(section))

    updater["section0"].name = "section1"

    expected = """\
    [section1] # section comment
    key = value
    """
    assert str(updater) == dedent(expected)
    assert updater["section1"].raw_comment == " # section comment"

    updater["section1"].raw_comment = " # new comment"
    expected = """\
    [section1] # new comment
    key = value
    """
    assert str(updater) == dedent(expected)


def test_setitem_detached_option():
    existing = """\
    [section0]
    option0 = 0
    option1 = # No value
    """

    template1 = """\
    [section1]
    option1 = 1
    """

    target = ConfigUpdater()
    target.read_string(dedent(existing))

    source1 = ConfigUpdater()
    source1.read_string(dedent(template1))

    option1 = source1["section1"]["option1"].detach()
    target["section0"]["option1"] = option1
    assert target["section0"]["option1"].value == "1"


def test_comments_in_multiline_values():
    indented_case = """\
    [header]
    install_requires =
        importlib-metadata; python_version<"3.8"
        #Adding some comments here that are perfectly valid.
        some-other-dependency
    """
    unindented_case = """\
    [header]
    install_requires =
        importlib-metadata; python_version<"3.8"
    #Adding some comments here that are perfectly valid.
        some-other-dependency
    """
    error_case_if_no_value = """\
    [header]
    install_requires =
        importlib-metadata; python_version<"3.8"
    #Adding some comments here that are perfectly valid.
    some-other-dependency
    """
    parser = ConfigParser(allow_no_value=False)
    updater = ConfigUpdater(allow_no_value=False)

    parser.read_string(dedent(indented_case))
    updater.read_string(dedent(indented_case))
    assert str(updater) == dedent(indented_case)

    parser.read_string(dedent(unindented_case))
    updater.read_string(dedent(unindented_case))
    assert str(updater) == dedent(unindented_case)

    with pytest.raises(configparser.ParsingError):
        parser.read_string(dedent(error_case_if_no_value))
    with pytest.raises(ParsingError):
        updater.read_string(dedent(error_case_if_no_value))

    parser = ConfigParser(allow_no_value=True)
    updater = ConfigUpdater(allow_no_value=True)

    parser.read_string(dedent(indented_case))
    updater.read_string(dedent(indented_case))
    assert str(updater) == dedent(indented_case)

    parser.read_string(dedent(unindented_case))
    updater.read_string(dedent(unindented_case))
    assert str(updater) == dedent(unindented_case)

    parser.read_string(dedent(error_case_if_no_value))
    updater.read_string(dedent(error_case_if_no_value))
    assert str(updater) == dedent(error_case_if_no_value)


def test_modify_multiline_value():
    ml_value = """\
    [metadata]
    classifiers =
        Development Status :: 3 - Alpha
        #Development Status :: 4 - Beta
        #Development Status :: 5 - Production/Stable
        Programming Language :: Python :: 3 :: Only
        Programming Language :: Python :: 3
        Programming Language :: Python :: Implementation :: CPython
        Programming Language :: Python :: Implementation :: PyPy
        License :: OSI Approved :: MIT License
    """
    ml_value = dedent(ml_value)
    updater = ConfigUpdater()
    updater.read_string(ml_value)

    with pytest.raises(AssignMultilineValueError):
        updater["metadata"]["classifiers"].value += "License :: OSI Approved :: BSD"

    new_classifiers = updater["metadata"]["classifiers"].value.strip().split("\n")
    new_classifiers += ["Topic :: Utilities"]
    updater["metadata"]["classifiers"].set_values(new_classifiers)
    ml_value += "    Topic :: Utilities\n"
    assert str(updater) == ml_value

    updater["metadata"]["classifiers"].value = "new_value"
    updater["metadata"]["classifiers"].value += "_and_more"
    expected = """\
    [metadata]
    classifiers = new_value_and_more
    """
    assert str(updater) == dedent(expected)


def test_append_to_multi_line_value():
    cfg = """\
    [metadata]
    classifiers =
        Development Status :: 3 - Alpha
        #Development Status :: 4 - Beta
        #Development Status :: 5 - Production/Stable
        Programming Language :: Python :: 3 :: Only
        Programming Language :: Python :: 3
        Programming Language :: Python :: Implementation :: CPython
        Programming Language :: Python :: Implementation :: PyPy
        License :: OSI Approved :: MIT License
    """
    cfg = dedent(cfg)
    updater = ConfigUpdater()
    updater.read_string(cfg)
    updater["metadata"]["classifiers"].append("Topic :: Utilities")
    cfg += "    Topic :: Utilities\n"
    assert str(updater) == cfg


def test_append_to_comma_separated_value():
    cfg = """\
    [metadata]
    classifiers = terminal, developers
    """
    cfg = dedent(cfg)
    updater = ConfigUpdater()
    updater.read_string(cfg)
    option = updater["metadata"]["classifiers"]
    option.append("gui", separator=", ")

    cfg_out = """\
    [metadata]
    classifiers = terminal, developers, gui
    """
    cfg_out = dedent(cfg_out)
    assert str(updater) == cfg_out


def test_append_to_no_multi_line_value():
    cfg_in = """\
    [main]
    key = value1
    """
    cfg_out = """\
    [main]
    key =
        value1
        new_value
    """
    cfg_in, cfg_out = dedent(cfg_in), dedent(cfg_out)
    updater = ConfigUpdater()
    updater.read_string(cfg_in)
    updater["main"]["key"].append("new_value")

    target = ConfigUpdater()
    target.read_string(cfg_out)
    assert str(updater) == str(target)

    updater["main"]["key"].append("new_value2")
    cfg_out += "    new_value2\n"
    assert str(updater) == cfg_out


def test_as_list_multi_line():
    cfg = """\
    [metadata]
    no_value
    single_value = 1
    classifiers =
        Development Status :: 3 - Alpha
        #Development Status :: 4 - Beta
    """
    cfg = dedent(cfg)
    updater = ConfigUpdater(allow_no_value=True)
    updater.read_string(cfg)

    meta = updater["metadata"]
    assert meta["no_value"].as_list() == []
    assert meta["single_value"].as_list() == ["1"]
    assert meta["classifiers"].as_list() == [
        "Development Status :: 3 - Alpha",
        "#Development Status :: 4 - Beta",
    ]


def test_as_list_comma_separated():
    cfg = """\
    [metadata]
    classifiers = terminal, developers
    """
    cfg = dedent(cfg)
    updater = ConfigUpdater()
    updater.read_string(cfg)
    option = updater["metadata"]["classifiers"]
    assert option.as_list(",") == ["terminal", "developers"]


def test_indented_section():
    cfg = """\
    [section0]

        [section1]
            key = value
    """

    updater = ConfigUpdater()
    updater.read_string(dedent(cfg))

    assert str(updater) == dedent(cfg)


def test_section_without_end_of_line():
    cfg = """\
    [section]"""

    updater = ConfigUpdater()
    updater.read_string(dedent(cfg))

    assert str(updater) == dedent(cfg)



import sys

if sys.version_info[:2] >= (3, 8):  # pragma: no cover
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version
else:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "ConfigUpdater"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

# import everything and rely on __ALL__
from .configupdater import *  # noqa
from .parser import *  # noqa



from copy import deepcopy
from textwrap import dedent

import pytest

from configupdater.block import NotAttachedError
from configupdater.configupdater import ConfigUpdater
from configupdater.option import NoneValueDisallowed, Option
from configupdater.parser import Parser


def test_deepcopy():
    example = """\
    [options.extras_require]
    testing =   # Add here test requirements (used by tox)
        sphinx  # required for system tests
        flake8  # required for system tests
    """
    doc = Parser().read_string(dedent(example))
    section = doc["options.extras_require"]
    option = section["testing"]
    assert option.container is section

    clone = deepcopy(option)

    assert str(clone) == str(option)
    assert option.container is section
    with pytest.raises(NotAttachedError):
        assert clone.container is None  # copies should always be created detached

    # Make sure no side effects are felt by the original when the copy is modified
    clone.value = ""
    assert str(clone) != str(option)
    assert str(doc) == dedent(example)


def test_str_for_none_value():
    assert str(Option("namespace")) == "namespace\n"  # orphan option

    cfg = ConfigUpdater(allow_no_value=True).read_string("[pyscaffold]")
    cfg.set("pyscaffold", "namespace")
    assert str(cfg["pyscaffold"]["namespace"]) == "namespace\n"

    cfg = ConfigUpdater(allow_no_value=False).read_string("[pyscaffold]")
    cfg.set("pyscaffold", "namespace")
    with pytest.warns(NoneValueDisallowed):
        assert str(cfg["pyscaffold"]["namespace"]) == ""


def test_str_for_set_values():
    opt = Option("opt")
    opt.set_values("1234", separator=", ")
    assert str(opt) == "opt = 1, 2, 3, 4\n"



"""Together with :mod:`~configupdater.container` this module forms the basis of
the class hierarchy in **ConfigUpdater**.

The :class:`Block` is the parent class of everything that can be nested inside a
configuration file, e.g. comments, sections, options and even sequences of white space.
"""
import sys
from abc import ABC
from copy import deepcopy
from inspect import cleandoc
from typing import TYPE_CHECKING, Optional, TypeVar, Union, cast

if sys.version_info[:2] >= (3, 9):  # pragma: no cover
    List = list
else:  # pragma: no cover
    from typing import List

if TYPE_CHECKING:
    from .builder import BlockBuilder
    from .container import Container

T = TypeVar("T")
B = TypeVar("B", bound="Block")


def _short_repr(block) -> str:
    if isinstance(block, str):
        return block
    name = getattr(block, "raw_key", None) or getattr(block, "name", None)
    name = f" {name!r}" if name else ""
    return f"<{block.__class__.__name__}{name}>"


class NotAttachedError(Exception):
    """{block} is not attached to a container yet. Try to insert it first."""

    def __init__(self, block: Union[str, "Block"] = "The block"):
        doc = cleandoc(cast(str, self.__class__.__doc__))
        msg = doc.format(block=_short_repr(block))
        super().__init__(msg)


class AlreadyAttachedError(Exception):
    """{block} has been already attached to a container.

    Try to remove it first using ``detach`` or create a copy using stdlib's
    ``copy.deepcopy``.
    """

    def __init__(self, block: Union[str, "Block"] = "The block"):
        doc = cleandoc(cast(str, self.__class__.__doc__))
        msg = doc.format(block=_short_repr(block))
        super().__init__(msg)


class AssignMultilineValueError(Exception):
    """Trying to assign a multi-line value to {block}.
    Use the ``set_values`` or ``append`` method to accomplish that.
    """

    def __init__(self, block: Union[str, "Block"] = "The block"):
        doc = cleandoc(cast(str, self.__class__.__doc__))
        msg = doc.format(block=_short_repr(block))
        super().__init__(msg)


class Block(ABC):
    """Abstract Block type holding lines

    Block objects hold original lines from the configuration file and hold
    a reference to a container wherein the object resides.

    The type variable ``T`` is a reference for the type of the sibling blocks
    inside the container.
    """

    def __init__(self, container: Optional["Container"] = None):
        self._container = container
        self._lines: List[str] = []
        self._updated = False

    def __str__(self) -> str:
        return "".join(self._lines)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {str(self)!r}>"

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return str(self) == str(other)
        else:
            return False

    def __deepcopy__(self: B, memo: dict) -> B:
        clone = self._instantiate_copy()
        clone._lines = deepcopy(self._lines, memo)
        clone._updated = self._updated
        memo[id(self)] = clone
        return clone

    def _instantiate_copy(self: B) -> B:
        """Auxiliary method that allows subclasses calling ``__deepcopy__``"""
        return self.__class__(container=None)  # allow overwrite for different init args
        # ^  A fresh copy should always be made detached from any container

    def add_line(self: B, line: str) -> B:
        """PRIVATE: this function is not part of the public API of Block.
        It is only used internally by other classes of the package during parsing.

        Add a line to the current block

        Args:
            line (str): one line to add
        """
        self._lines.append(line)
        return self

    @property
    def lines(self) -> List[str]:
        return self._lines

    @property
    def container(self) -> "Container":
        """Container holding the block"""
        if self._container is None:
            raise NotAttachedError(self)
        return self._container

    @property
    def container_idx(self: B) -> int:
        """Index of the block within its container"""
        return self.container.structure.index(self)

    @property
    def updated(self) -> bool:
        """True if the option was changed/updated, otherwise False"""
        # if no lines were added, treat it as updated since we added it
        return self._updated or not self._lines

    def _builder(self, idx: int) -> "BlockBuilder":
        from .builder import BlockBuilder

        return BlockBuilder(self.container, idx)

    @property
    def add_before(self) -> "BlockBuilder":
        """Block builder inserting a new block before the current block"""
        return self._builder(self.container_idx)

    @property
    def add_after(self) -> "BlockBuilder":
        """Block builder inserting a new block after the current block"""
        return self._builder(self.container_idx + 1)

    @property
    def next_block(self) -> Optional["Block"]:
        """Next block in the current container"""
        idx = self.container_idx + 1
        if idx < len(self.container):
            return self.container.structure[idx]
        else:
            return None

    @property
    def previous_block(self) -> Optional["Block"]:
        """Previous block in the current container"""
        idx = self.container_idx - 1
        if idx >= 0:
            return self.container.structure[idx]
        else:
            return None

    def detach(self: B) -> B:
        """Remove and return this block from container"""
        self.container._remove_block(self.container_idx)
        self._container = None
        return self

    def has_container(self) -> bool:
        """Checks if this block has a container attached"""
        return not (self._container is None)

    def attach(self: B, container: "Container") -> B:
        """PRIVATE: Don't use this as a user!

        Rather use `add_*` or the bracket notation
        """
        if self._container is not None and self._container is not container:
            raise AlreadyAttachedError(self)
        self._container = container
        return self


class Comment(Block):
    """Comment block"""


class Space(Block):
    """Vertical space block of new lines"""



from copy import deepcopy
from textwrap import dedent

import pytest

from configupdater.block import AlreadyAttachedError, NotAttachedError
from configupdater.parser import Parser


def test_set():
    example = """\
    [options.extras_require]
    testing =   # Add here test requirements (used by tox)
        sphinx  # required for system tests
        flake8  # required for system tests
    """
    doc = Parser().read_string(dedent(example))
    section = doc["options.extras_require"]

    section.set("all", "pyscaffold")
    assert section["all"].value == "pyscaffold"

    section.set("testing", ["pyscaffoldext-markdown", "rst-to-myst"])
    assert section["testing"].value == "\n    pyscaffoldext-markdown\n    rst-to-myst"


def test_deepcopy():
    example = """\
    [options.extras_require]
    testing =   # Add here test requirements (used by tox)
        sphinx  # required for system tests
        flake8  # required for system tests
    """
    doc = Parser().read_string(dedent(example))
    other = Parser().read_string("")
    section = doc["options.extras_require"]
    option = section["testing"]
    assert option.container is section

    clone = deepcopy(section)
    with pytest.raises(NotAttachedError):  # copies should always be created detached
        assert clone.container is None

    other.add_section(clone)  # needed to be able to modify section
    assert clone.container is other

    assert str(clone) == str(section)
    assert section.container is doc

    # Make sure no side effects are felt by the original when the copy is modified
    # and vice-versa
    clone["testing"] = ""
    assert str(clone) != str(section)
    assert str(doc) == dedent(example)
    clone["testing"].add_before.option("extra_option", "extra_value")
    assert "extra_option" in clone
    assert "extra_option" not in section
    assert clone["extra_option"].container is clone

    section["testing"].add_before.option("other_extra_option", "other_extra_value")
    assert "other_extra_option" in section
    assert "other_extra_option" not in clone
    assert section["other_extra_option"].container is section

    section.add_after.comment("# new comment")
    assert "# new comment" in str(doc)
    assert "# new comment" not in str(other)

    clone.add_before.comment("# other comment")
    assert "# other comment" in str(other)
    assert "# other comment" not in str(doc)


def test_clear_error_message():
    # Make sure the error messages specify the exact object
    example = """\
    [options.extras_require]
    testing =   # Add here test requirements (used by tox)
        sphinx  # required for system tests
        flake8  # required for system tests
    """
    doc = Parser().read_string(dedent(example))
    section = doc["options.extras_require"]
    clone = deepcopy(section)
    with pytest.raises(NotAttachedError) as ex:
        clone["testing"] = ""
    assert "<Section 'options.extras_require'>" in str(ex.value)

    with pytest.raises(AlreadyAttachedError) as ex:
        section["testing"] = next(clone.iter_options())
    assert "<Option 'testing'>" in str(ex.value)



"""Together with :mod:`~configupdater.block` this module forms the basis of
the class hierarchy in **ConfigUpdater**.

The :class:`Container` is the parent class of everything that can contain configuration
blocks, e.g. a section or the entire file itself.
"""
import sys
from abc import ABC
from copy import deepcopy
from textwrap import indent
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

if sys.version_info[:2] >= (3, 9):  # pragma: no cover
    from collections.abc import Iterator

    List = list
else:  # pragma: no cover
    from typing import Iterator, List

if TYPE_CHECKING:
    from .block import Block  # noqa

T = TypeVar("T", bound="Block")
C = TypeVar("C", bound="Container")


class Container(ABC, Generic[T]):
    """Abstract Mixin Class describing a container that holds blocks of type ``T``"""

    def __init__(self):
        self._structure: List[T] = []

    def _repr_blocks(self) -> str:
        blocks = "\n".join(repr(block) for block in self._structure)
        blocks = indent(blocks, " " * 4)
        return f"[\n{blocks.rstrip()}\n]" if blocks.strip() else "[]"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._repr_blocks()}>"

    def __deepcopy__(self: C, memo: dict) -> C:
        clone = self._instantiate_copy()
        memo[id(self)] = clone
        return clone._copy_structure(self._structure, memo)

    def _copy_structure(self: C, structure: List[T], memo: dict) -> C:
        """``__deepcopy__`` auxiliary method also useful with multi-inheritance"""
        self._structure = [b.attach(self) for b in deepcopy(structure, memo)]
        return self

    def _instantiate_copy(self: C) -> C:
        """Auxiliary method that allows subclasses calling ``__deepcopy__``"""
        return self.__class__()  # allow overwrite for different init args

    @property
    def structure(self) -> List[T]:
        return self._structure

    @property
    def first_block(self) -> Optional[T]:
        if self._structure:
            return self._structure[0]
        else:
            return None

    @property
    def last_block(self) -> Optional[T]:
        if self._structure:
            return self._structure[-1]
        else:
            return None

    def _remove_block(self: C, idx: int) -> C:
        """Remove block at index idx within container

        Use `.container_idx` of a block to get the index.
        Not meant for users, rather use block.detach() instead!
        """
        del self._structure[idx]
        return self

    def iter_blocks(self) -> Iterator[T]:
        """Iterate over all blocks inside container."""
        return iter(self._structure)

    def __len__(self) -> int:
        """Number of blocks in container"""
        return len(self._structure)

    def append(self: C, block: T) -> C:
        self._structure.append(block)
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for configupdater.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
from __future__ import absolute_import, division, print_function

import inspect
import os

import pytest


@pytest.fixture
def setup_cfg_path():
    filepath = inspect.getfile(inspect.currentframe())
    filedir = os.path.dirname(os.path.abspath(filepath))
    return os.path.join(filedir, "test_setup.cfg")


@pytest.fixture
def setup_cfg(setup_cfg_path):
    with open(setup_cfg_path) as fh:
        return fh.read()



"""This module focus in the top level data layer API of ConfigUpdater, i.e.
how to access and modify the sections of the configurations.

Differently from :mod:`configparser`, the different aspects of the ConfigUpdater API are
split between several modules.
"""
import sys
from configparser import (
    ConfigParser,
    DuplicateSectionError,
    NoOptionError,
    NoSectionError,
)
from enum import Enum
from typing import Optional, Tuple, TypeVar, Union, overload

if sys.version_info[:2] >= (3, 9):  # pragma: no cover
    from collections.abc import Iterable, Iterator, MutableMapping

    List = list
    Dict = dict
else:  # pragma: no cover
    from typing import Dict, Iterable, Iterator, List, MutableMapping

from .block import Comment, Space
from .container import Container
from .option import Option
from .section import Section

# Used in parser getters to indicate the default behaviour when a specific
# option is not found it to raise an exception. Created to enable 'None' as
# a valid fallback value.
_UniqueValues = Enum("UniqueValues", "_UNSET")
_UNSET = _UniqueValues._UNSET

T = TypeVar("T")
D = TypeVar("D", bound="Document")

ConfigContent = Union["Section", "Comment", "Space"]
Value = Union["Option", str]


class Document(Container[ConfigContent], MutableMapping[str, Section]):
    """Access to the data manipulation API from **ConfigUpdater**.

    A ``Document`` object tries to implement a familiar *dict-like* interface,
    via :class:`MutableMapping`. However, it also tries to be as compatible as possible
    with the stdlib's :class:`~configparser.ConfigParser`. This means that there are a
    few methods that will work differently from what users familiar with *dict-like*
    interfaces would expect. The most notable example is :meth:`get`.

    A important difference between ConfigUpdater's ``Document`` model and
    :class:`~configparser.ConfigParser` is the behaviour of the :class:`Section`
    objects.
    If we represent the type of a *dict-like* (or :class:`MutableMapping`) object by
    ``M[K, V]``, where ``K`` is the type of its keys and ``V`` is the type of the
    associated values, ConfigUpdater's sections would be equivalent to
    ``M[str, Option]``, while ConfigParser's would be ``M[str, str]``.

    This means that when you try to access a key inside a section in ConfigUpdater, you
    are going to receive a :class:`Option` object, not its value.
    To access the value of the option you need to call :class:`Option.value`.
    """

    def _get_section_idx(self, name: str) -> int:
        return next(
            i
            for i, entry in enumerate(self._structure)
            if isinstance(entry, Section) and entry.name == name
        )

    def optionxform(self, optionstr) -> str:
        """Converts an option key for unification

        By default it uses :meth:`str.lower`, which means that ConfigUpdater will
        compare options in a case insensitive way.

        This implementation mimics ConfigParser API, and can be configured as described
        in :meth:`configparser.ConfigParser.optionxform`.

        Args:
             optionstr (str): key name

        Returns:
            str: unified option name
        """
        return optionstr.lower()

    def validate_format(self, **kwargs):
        """Call ConfigParser to validate config

        Args:
            kwargs: are passed to :class:`configparser.ConfigParser`

        Raises:
            configparser.ParsingError: if syntax errors are found

        Returns:
            ``True`` when no error is found
        """
        kwargs.pop("space_around_delimiters", None)
        parser = ConfigParser(**kwargs)
        parser.read_string(str(self))
        return True

    def iter_sections(self) -> Iterator[Section]:
        """Iterate only over section blocks"""
        return (block for block in self._structure if isinstance(block, Section))

    def section_blocks(self) -> List[Section]:
        """Returns all section blocks

        Returns:
            list: list of :class:`Section` blocks
        """
        return list(self.iter_sections())

    def sections(self) -> List[str]:
        """Return a list of section names

        Returns:
            list: list of section names
        """
        return [section.name for section in self.iter_sections()]

    def __iter__(self) -> Iterator[str]:
        return (b.name for b in self.iter_blocks() if isinstance(b, Section))

    def __str__(self) -> str:
        return "".join(str(block) for block in self._structure)

    def __getitem__(self, key) -> Section:
        for section in self.section_blocks():
            if section.name == key:
                return section

        raise KeyError(f"No section `{key}` found", {"key": key})

    def __setitem__(self, key: str, value: Section):
        if not isinstance(value, Section):
            raise ValueError("Value must be of type Section!")
        if isinstance(key, str) and key in self:
            idx = self._get_section_idx(key)
            self.__getitem__(key).detach()
            value.attach(self)
            self._structure.insert(idx, value)
        else:
            if value.name != key:
                raise ValueError(
                    f"Set key `{key}` does not equal option key `{value.name}`"
                )
            self.add_section(value)

    def __delitem__(self, key: str):
        if not self.remove_section(key):
            raise KeyError(f"No section `{key}` found", {"key": key})

    # MutableMapping[str, Section] for some reason accepts key: object
    # it actually doesn't matter for the implementation, so we omit the typing
    def __contains__(self, key) -> bool:
        """Returns whether the given section exists.

        Args:
            key (str): name of section

        Returns:
            bool: wether the section exists
        """
        return next((True for s in self.iter_sections() if s.name == key), False)

    has_section = __contains__

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self._structure == other._structure
        else:
            return False

    def clear(self):
        for block in self._structure:
            block.detach()
        self._structure.clear()

    def add_section(self, section: Union[str, Section]):
        """Create a new section in the configuration.

        Raise DuplicateSectionError if a section by the specified name
        already exists. Raise ValueError if name is DEFAULT.

        Args:
            section (str or :class:`Section`): name or Section type
        """
        if isinstance(section, str):
            # create a new section
            section_obj = Section(section)
        elif isinstance(section, Section):
            section_obj = section
        else:
            raise ValueError("Parameter must be a string or Section type!")

        if self.has_section(section_obj.name):
            raise DuplicateSectionError(section_obj.name)

        section_obj.attach(self)
        self._structure.append(section_obj)

    def options(self, section: str) -> List[str]:
        """Returns list of configuration options for the named section.

        Args:
            section (str): name of section

        Returns:
            list: list of option names
        """
        if not self.has_section(section):
            raise NoSectionError(section) from None
        return self.__getitem__(section).options()

    # The following is a pragmatic violation of Liskov substitution principle:
    # As dicts, Mappings should have get(self, key: str, default: T) -> T
    # but ConfigParser overwrites it and uses the function to offer a different
    # functionality
    @overload  # type: ignore[override]
    def get(self, section: str, option: str) -> Option:
        ...

    @overload
    def get(self, section: str, option: str, fallback: T) -> Union[Option, T]:  # noqa
        ...

    def get(self, section, option, fallback=_UNSET):  # noqa
        """Gets an option value for a given section.

        Warning:
            Please notice this method works differently from what is expected of
            :meth:`MutableMapping.get` (or :meth:`dict.get`).
            Similarly to :meth:`configparser.ConfigParser.get`, will take least 2
            arguments, and the second argument does not correspond to a default value.

            This happens because this function is not designed to return a
            :obj:`Section` of the :obj:`ConfigUpdater` document, but instead a nested
            :obj:`Option`.

            See :meth:`get_section`, if instead, you want to retrieve a :obj:`Section`.

        Args:
            section (str): section name
            option (str): option name
            fallback: if the key is not found and fallback is provided, it will
                be returned. ``None`` is a valid fallback value.

        Raises:
            :class:`NoSectionError`: if ``section`` cannot be found
            :class:`NoOptionError`: if the option cannot be found and no ``fallback``
                was given

        Returns:
            :class:`Option`: Option object holding key/value pair
        """
        section_obj = self.get_section(section, _UNSET)
        if section_obj is _UNSET:
            raise NoSectionError(section) from None

        option = self.optionxform(option)
        value = section_obj.get(option, fallback)
        # ^  we checked section_obj against _UNSET, so we are sure about its type

        if value is _UNSET:
            raise NoOptionError(option, section)

        return value

    @overload
    def get_section(self, name: str) -> Optional[Section]:
        ...

    @overload
    def get_section(self, name: str, default: T) -> Union[Section, T]:
        ...

    def get_section(self, name, default=None):
        """This method works similarly to :meth:`dict.get`, and allows you
        to retrieve an entire section by its name, or provide a ``default`` value in
        case it cannot be found.
        """
        return next((s for s in self.iter_sections() if s.name == name), default)

    # The following is a pragmatic violation of Liskov substitution principle
    # For some reason MutableMapping.items return a Set-like object
    # but we want to preserve ordering
    @overload  # type: ignore[override]
    def items(self) -> List[Tuple[str, Section]]:
        ...

    @overload
    def items(self, section: str) -> List[Tuple[str, Option]]:  # noqa
        ...

    def items(self, section=_UNSET):  # noqa
        """Return a list of (name, value) tuples for options or sections.

        If section is given, return a list of tuples with (name, value) for
        each option in the section. Otherwise, return a list of tuples with
        (section_name, section_type) for each section.

        Args:
            section (str): optional section name, default UNSET

        Returns:
            list: list of :class:`Section` or :class:`Option` objects
        """
        if section is _UNSET:
            return [(sect.name, sect) for sect in self.iter_sections()]

        section = self.__getitem__(section)
        return [(opt.key, opt) for opt in section.iter_options()]

    def has_option(self, section: str, option: str) -> bool:
        """Checks for the existence of a given option in a given section.

        Args:
            section (str): name of section
            option (str): name of option

        Returns:
            bool: whether the option exists in the given section
        """
        key = self.optionxform(option)
        return key in self.get_section(section, {})

    def set(
        self: D,
        section: str,
        option: str,
        value: Union[None, str, Iterable[str]] = None,
    ) -> D:
        """Set an option.

        Args:
            section: section name
            option: option name
            value: value, default None
        """
        try:
            section_obj = self.__getitem__(section)
        except KeyError:
            raise NoSectionError(section) from None
        option = self.optionxform(option)
        section_obj.set(option, value)
        return self

    def remove_option(self, section: str, option: str) -> bool:
        """Remove an option.

        Args:
            section (str): section name
            option (str): option name

        Returns:
            bool: whether the option was actually removed
        """
        try:
            section_obj = self.__getitem__(section)
        except KeyError:
            raise NoSectionError(section) from None
        option = self.optionxform(option)
        existed = option in section_obj.options()
        if existed:
            del section_obj[option]
        return existed

    def remove_section(self, name: str) -> bool:
        """Remove a file section.

        Args:
            name: name of the section

        Returns:
            bool: whether the section was actually removed
        """
        try:
            idx = self._get_section_idx(name)
            del self._structure[idx]
            return True
        except StopIteration:
            return False

    def to_dict(self) -> Dict[str, Dict[str, Optional[str]]]:
        """Transform to dictionary

        Returns:
            dict: dictionary with same content
        """
        return {sect.name: sect.to_dict() for sect in self.iter_sections()}



import pytest

from configupdater.parser import Parser


def test_syntax_options():
    parser = Parser(allow_no_value=True)
    assert parser.syntax_options["allow_no_value"] is True


def test_syntax_options_read_only():
    parser = Parser(allow_no_value=False)
    with pytest.raises(Exception) as exc_info:
        parser.syntax_options["allow_no_value"] = True
    assert "assignment" in str(exc_info.value)
    assert parser.syntax_options["allow_no_value"] is False



"""Options are the ultimate mean of configuration inside a configuration value.

They are always associated with a :attr:`~Option.key` (or the name of the configuration
parameter) and a :attr:`~Option.value`.

Options can also have multi-line values that are usually interpreted as a list of
values.

When editing configuration files with ConfigUpdater, a handy way of setting a multi-line
(or comma separated value) for an specific option is to use the
:meth:`~Option.set_values` method.
"""
import sys
import warnings
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, cast

if sys.version_info[:2] >= (3, 9):  # pragma: no cover
    from collections.abc import Iterable

    List = list
    Dict = dict
else:  # pragma: no cover
    from typing import Iterable, List

if TYPE_CHECKING:
    from .section import Section
    from .document import Document

from .block import AssignMultilineValueError, Block

Value = Union["Option", str]
T = TypeVar("T", bound="Option")


def is_multi_line(value: Any) -> bool:
    if isinstance(value, str):
        return "\n" in value
    else:
        return False


class NoneValueDisallowed(SyntaxWarning):
    """Cannot represent <{option} = None>, it will be converted to <{option} = ''>.
    Please use ``allow_no_value=True`` with ``ConfigUpdater``.
    """

    @classmethod
    def warn(cls, option):
        warnings.warn(cls.__doc__.format(option=option), cls, stacklevel=2)


class Option(Block):
    """Option block holding a key/value pair.

    Attributes:
        key (str): name of the key
        value (str): stored value
        updated (bool): indicates name change or a new section
    """

    def __init__(
        self,
        key: str,
        value: Optional[str] = None,
        container: Optional["Section"] = None,
        delimiter: str = "=",
        space_around_delimiters: bool = True,
        line: Optional[str] = None,
    ):
        super().__init__(container=container)
        self._key = key
        self._values: List[Optional[str]] = [] if value is None else [value]
        self._value_is_none = value is None
        self._delimiter = delimiter
        self._value: Optional[str] = None  # will be filled after join_multiline_value
        self._updated = False
        self._multiline_value_joined = False
        self._space_around_delimiters = space_around_delimiters
        if line:
            super().add_line(line)
        if value is not None:
            self._set_value(value)

    def add_value(self, value: Optional[str]):
        """PRIVATE: this function is not part of the public API of Option.
        It is only used internally by other classes of the package during parsing.
        """
        self._value_is_none = value is None
        self._values.append(value)

    def add_line(self, line: str):
        """PRIVATE: this function is not part of the public API of Option.
        It is only used internally by other classes of the package during parsing.
        """
        super().add_line(line)
        self.add_value(line.strip())

    def _join_multiline_value(self):
        if not self._multiline_value_joined and not self._value_is_none:
            # do what `_join_multiline_value` in ConfigParser would do
            self._value = "\n".join(self._values).rstrip()
            self._multiline_value_joined = True

    def __str__(self) -> str:
        if not self.updated:
            return super().__str__()

        document = self._document()
        opts = getattr(document, "syntax_options", None) or {}
        value = self._value

        if value is None:
            if document is None or opts.get("allow_no_value"):
                return f"{self._key}\n"
            NoneValueDisallowed.warn(self._key)
            return ""

        space = self._space_around_delimiters or opts.get("space_around_delimiters")
        if space:
            # no space is needed if we use multi-line arguments
            suffix = "" if str(value).startswith("\n") else " "
            delim = f" {self._delimiter}{suffix}"
        else:
            delim = self._delimiter
        return f"{self._key}{delim}{value}\n"

    def __repr__(self) -> str:
        return f"<Option: {self._key} = {self.value!r}>"

    def _instantiate_copy(self: T) -> T:
        """Will be called by :meth:`Block.__deepcopy__`"""
        self._join_multiline_value()
        return self.__class__(
            self._key,
            self._value,
            container=None,
            delimiter=self._delimiter,
            space_around_delimiters=self._space_around_delimiters,
        )

    def _document(self) -> Optional["Document"]:
        if self._container is None:
            return None
        return self._container._container  # type: ignore

    @property
    def section(self) -> "Section":
        return cast("Section", self.container)

    @property
    def key(self) -> str:
        """Key string associated with the option.

        Please notice that the option key is normalized with
        :meth:`~configupdater.document.Document.optionxform`.

        When the option object is :obj:`detached <configupdater.block.Block.detach>`,
        this method will raise a :obj:`NotAttachedError`.
        """
        return self.section.document.optionxform(self._key)

    @key.setter
    def key(self, value: str):
        self._join_multiline_value()
        self._key = value
        self._updated = True

    @property
    def raw_key(self) -> str:
        """Equivalent to :obj:`key`, but before applying :meth:`optionxform`."""
        return self._key

    @property
    def value(self) -> Optional[str]:
        self._join_multiline_value()
        return self._value

    @value.setter
    def value(self, value: str):
        if is_multi_line(value):
            raise AssignMultilineValueError(self)
        self._set_value(value)

    def _set_value(self, value: str):
        self._updated = True
        self._multiline_value_joined = True
        self._value = value
        self._values = [value]

    def as_list(self, separator="\n") -> List[str]:
        """Returns the (multi-line/element) value as a list

        Empty list if value is None, single-element list for a one-element
        value and an element for each line in a multi-element value.

        Args:
            separator (str): separator for values, default: line separator
        """
        if self._value_is_none:
            return []
        else:
            return [v.strip() for v in cast(str, self.value).strip().split(separator)]

    def append(self, value: str, **kwargs) -> "Option":
        """Append a value to a mult-line value

        Args:
            value (str): value
            kwargs: keyword arguments for `set_values`
        """
        sep = kwargs.get("separator", None)
        if sep is None:
            new_values = self.as_list()
        else:
            new_values = self.as_list(separator=sep)

        new_values.append(value)
        self.set_values(new_values, **kwargs)
        return self

    def set_values(self, values: Iterable[str], separator="\n", indent=4 * " "):
        """Sets the value to a given list of options, e.g. multi-line values

        Args:
            values (iterable): sequence of values
            separator (str): separator for values, default: line separator
            indent (str): indentation depth in case of line separator
        """
        values = list(values).copy()
        self._updated = True
        self._multiline_value_joined = True
        self._values = cast(List[Optional[str]], values)
        if "\n" in separator:
            values = [""] + values
            separator = separator + indent
        self._value = separator.join(values)



from textwrap import dedent

import pytest

from configupdater import ConfigUpdater


def test_stdlib_docs_compatibility():
    """Make sure optionxform works as documented in the stdlib docs (ConfigParser)"""
    config = """\
    [Section1]
    Key = Value

    [Section2]
    AnotherKey = Value
    """

    typical = ConfigUpdater()
    typical.read_string(dedent(config))
    assert list(typical["Section1"].keys()) == ["key"]
    assert list(typical["Section2"].keys()) == ["anotherkey"]

    custom = ConfigUpdater()
    custom.optionxform = lambda option: option
    custom.read_string(dedent(config))
    assert list(custom["Section1"].keys()) == ["Key"]
    assert list(custom["Section2"].keys()) == ["AnotherKey"]

    other = ConfigUpdater()
    other.optionxform = str
    other.read_string(dedent(config))
    assert list(other["Section1"].keys()) == ["Key"]
    assert list(other["Section2"].keys()) == ["AnotherKey"]


def test_custom():
    config = """\
    [section1]
    k_e_y = value
    """

    custom = ConfigUpdater()
    custom.optionxform = lambda option: option.replace("_", "")
    custom.read_string(dedent(config))
    assert list(custom["section1"].keys()) == ["key"]
    assert custom["section1"]["key"].value == "value"
    assert custom["section1"]["key"].raw_key == "k_e_y"


def test_section_contains():
    config = """\
    [section1]
    key = value
    """

    lowercase = ConfigUpdater()
    lowercase.read_string(dedent(config).lower())

    uppercase = ConfigUpdater()
    uppercase.optionxform = str
    uppercase.read_string(dedent(config).upper())

    assert "KEY" in uppercase["SECTION1"]
    assert "key" not in uppercase["SECTION1"]

    assert "key" in lowercase["section1"]
    assert "KEY" not in lowercase["section1"]


def test_section_setitem():
    cfg = ConfigUpdater()
    cfg.optionxform = str.upper
    cfg.read_string("[section1]\nOTHERKEY = 0")

    assert "KEY" not in cfg["section1"]
    cfg["section1"]["key"] = "value"
    assert "KEY" in cfg["section1"]
    assert cfg["section1"]["KEY"].value == "value"

    cfg["section1"]["key"] = "42"
    assert cfg["section1"]["KEY"].value == "42"
    assert cfg["section1"]["key"].value == "42"

    other = ConfigUpdater()
    other.optionxform = str.lower
    other.read_string("[section1]\nkEy = value")
    option = other["section1"]["key"].detach()

    with pytest.raises(ValueError):
        # otherkey exists in section1, but option is `key` instead of `otherkey`
        cfg["section1"]["otherkey"] = option
    with pytest.raises(ValueError):
        # anotherkey exists in section1 and option is `key` instead of `anotherkey`
        cfg["section1"]["anotherkey"] = option

    assert cfg["section1"]["key"].raw_key == "key"
    cfg["section1"]["key"] = option
    assert cfg["section1"]["key"].value == "value"
    assert cfg["section1"]["key"].key == "KEY"
    assert cfg["section1"]["key"].raw_key == "kEy"
